@@ -20,12 +20,13 @@
  *
  */
 
-const struct proxy_interface *device_wsc_get(const char *device_name);
+struct proxy_interface;
 
-char *device_wsc_family_arg_completion(const char *text, int state);
+char *device_arg_completion(const char *text, int state,
+				const struct command *commands,
+				const char *extra_interface);
 
-const struct proxy_interface *device_ap_get(const char *device_name);
-char *device_ap_family_arg_completion(const char *text, int state);
-
-const struct proxy_interface *device_ad_hoc_get(const char *device_name);
-char *device_ad_hoc_family_arg_completion(const char *text, int state);
+const struct proxy_interface *device_proxy_find_by_name(const char *name);
+const struct proxy_interface *device_proxy_find(const char *device_name,
+							const char *interface);
+const struct proxy_interface *device_get_default(void);
