@@ -24,14 +24,20 @@
 #include <config.h>
 #endif
 
+#if __STDC_VERSION__ <= 199409L
+#define _DEFAULT_SOURCE  /* for strto{u}ll() */
+#endif
+
+#define _GNU_SOURCE
 #include <stdio.h>
-#include <stdlib.h>
-#include <strings.h>
 #include <errno.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <stdlib.h>
+#include <string.h>
+#include <strings.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/mman.h>
 
 #include "util.h"
@@ -41,6 +47,7 @@
 #include "queue.h"
 #include "settings.h"
 #include "private.h"
+#include "missing.h"
 
 struct setting_data {
 	char *key;
@@ -64,7 +71,7 @@ static void setting_destroy(void *data)
 	struct setting_data *pair = data;
 
 	l_free(pair->key);
-	memset(pair->value, 0, strlen(pair->value));
+	explicit_bzero(pair->value, strlen(pair->value));
 	l_free(pair->value);
 	l_free(pair);
 }
@@ -135,6 +142,7 @@ static char *unescape_value(const char *value)
 			*n = '\\';
 			break;
 		default:
+			explicit_bzero(ret, n - ret);
 			l_free(ret);
 			return NULL;
 		}
@@ -347,7 +355,7 @@ static bool parse_keyvalue(struct l_settings *settings, const char *data,
 		return false;
 	}
 
-	if (parse_key(settings, data, equal - data, line) == false)
+	if (!parse_key(settings, data, equal - data, line))
 		return false;
 
 	equal += 1;
@@ -362,6 +370,7 @@ LIB_EXPORT bool l_settings_load_from_data(struct l_settings *settings,
 {
 	size_t pos = 0;
 	bool r = true;
+	bool has_group = false;
 	const char *eol;
 	size_t line = 1;
 	size_t line_len;
@@ -387,11 +396,17 @@ LIB_EXPORT bool l_settings_load_from_data(struct l_settings *settings,
 
 		line_len = eol - data - pos;
 
-		if (data[pos] == '[')
+		if (data[pos] == '[') {
 			r = parse_group(settings, data + pos, line_len, line);
-		else if (data[pos] != '#')
+			if (r)
+				has_group = true;
+		} else if (data[pos] != '#') {
+			if (!has_group)
+				return false;
+
 			r = parse_keyvalue(settings, data + pos, line_len,
 						line);
+		}
 
 		pos += line_len;
 	}
